/** 
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/
package org.activemq.ra;

import java.lang.reflect.Method;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.MessageProducer;
import javax.jms.ServerSession;
import javax.jms.Session;
import javax.resource.ResourceException;
import javax.resource.spi.endpoint.MessageEndpoint;
import javax.resource.spi.work.Work;
import javax.resource.spi.work.WorkEvent;
import javax.resource.spi.work.WorkException;
import javax.resource.spi.work.WorkListener;
import javax.resource.spi.work.WorkManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.activemq.ActiveMQSession;

/**
 * @version $Revision: 1.1.1.1 $
 */
public class ServerSessionImpl implements ServerSession, SessionAndProducer, Work {

    private static final Log log = LogFactory.getLog(ServerSessionImpl.class);
    public static final Method ON_MESSAGE_METHOD;

    static {
        try {
            ON_MESSAGE_METHOD = MessageListener.class.getMethod("onMessage", new Class[]{Message.class});
        }
        catch (Exception e) {
            throw new ExceptionInInitializerError(e);
        }
    }

    private ActiveMQSession session;
    private WorkManager workManager;
    private MessageEndpoint endpoint;
    private MessageProducer messageProducer;
    private final ServerSessionPoolImpl pool;

    private Object runControlMutex = new Object();
    private boolean runningFlag = false;

    public ServerSessionImpl(ServerSessionPoolImpl pool, ActiveMQSession session, WorkManager workManager, MessageEndpoint endpoint) throws JMSException {
        this.pool = pool;
        this.session = session;
        this.workManager = workManager;
        this.endpoint = endpoint;
        this.session.setMessageListener((MessageListener) endpoint);
    }

    public Session getSession() throws JMSException {
        return session;
    }

    public MessageProducer getMessageProducer() throws JMSException {
        if (messageProducer == null) {
            messageProducer = getSession().createProducer(null);
        }
        return messageProducer;
    }

    /**
     * @see javax.jms.ServerSession#start()
     */
    public void start() throws JMSException {

        log.debug("ServerSession started.");
        synchronized (runControlMutex) {
            if (runningFlag) {
                log.debug("ServerSession allready running.");
                return;
            }
            runningFlag = true;
        }

        // We get here because we need to start a async worker.
        log.debug("ServerSession queuing request for a run.");
        try {
            workManager.scheduleWork(this, WorkManager.INDEFINITE, null,
                    new WorkListener() {
                        //The work listener is useful only for debugging...
                        public void workAccepted(WorkEvent event) {
                            log.debug("Work accepted: " + event);
                        }

                        public void workRejected(WorkEvent event) {
                            log.debug("Work rejected: " + event);
                        }

                        public void workStarted(WorkEvent event) {
                            log.debug("Work started: " + event);
                        }

                        public void workCompleted(WorkEvent event) {
                            log.debug("Work completed: " + event);
                        }

                    });
        }
        catch (WorkException e) {
            throw (JMSException) new JMSException("Work could not be started: " + e).initCause(e);
        }
    }

    /**
     * @see java.lang.Runnable#run()
     */
    public void run() {
        
        log.debug(toString() + " begin run()"); 
        while (true) {
            log.debug(toString() + " top of run loop");
            log.debug("Running: " + this);
            
            synchronized (runControlMutex) {
                if( !session.hasUncomsumedMessages() ) {
                    runningFlag = false;
                    pool.returnToPool(this);
                    break;
                }                
            }
            
            try {
                SessionAndProducerHelper.register(this);                
                endpoint.beforeDelivery(ON_MESSAGE_METHOD);
                session.run();
            }
            catch (Throwable e) {
                log.info("Failed to process session: " + session + ". Reason: " + e, e);
            }            
            finally {
                try {
                    endpoint.afterDelivery();
                } catch (ResourceException e1) {
                    log.info("End point afterDelivery() failed. Reason: " + e1, e1);
                }
                SessionAndProducerHelper.unregister(this);
                
                // Sanitiy Check: If the local transaction has not been commited..
                // Commit it now.
                if( session.getTransactionContext().isInLocalTransaction() ) {
                    log.warn("Local transaction had not been commited.  Commiting now.");
                    try {
                        session.commit();
                    } catch (JMSException e) {
    	                log.info("Commit failed:", e);
                    }
                }
                
            }
            log.debug(toString() + " bottom of run loop");
        }
        log.debug(toString() + " end run()");
    }

    /**
     * @see javax.resource.spi.work.Work#release()
     */
    public void release() {
        log.debug("release called");
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString() {
        return "ServerSessionImpl[session=" + session + "]";
    }

}
